import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import IOResponseCommands
from iqrfpy.enums.message_types import IOMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.io.responses import GetResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': IOMessages.GET.value,
    'msgid': 'getTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 90,
    'result': {
        'ports': [206, 81, 24, 0, 8]
    },
    'dpa': b'\x00\x00\x09\x82\x00\x00\x00\x5a\xce\x51\x18\x00\x08'
}

data_ok_1: dict = {
    'mtype': IOMessages.GET.value,
    'msgid': 'getTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'ports': [206, 81, 24, 0, 8]
    },
    'dpa': b'\x00\x00\x09\x82\x02\x04\x00\x23\xce\x51\x18\x00\x08'
}

data_error: dict = {
    'mtype': IOMessages.GET.value,
    'msgid': 'getTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 4,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x09\x82\x04\x04\x04\x23'
}


class ReadResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, GetResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, GetResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, GetResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, GetResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, GetResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, GetResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.IO)
        with self.subTest():
            self.assertEqual(response.pcmd, IOResponseCommands.GET)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, IOMessages.GET)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    @parameterized.expand([
        ['from_dpa', data_ok['result']['ports'], GetResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['ports'], GetResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['ports'], GetResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['ports'], GetResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_data(self, _, data, response: GetResponse):
        self.assertEqual(response.port_data, data)

    @parameterized.expand([
        ['from_dpa_error', GetResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', GetResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_port_data_error(self, _, response: GetResponse):
        self.assertIsNone(response.port_data)



import time
from datetime import datetime
from iqrfpy.exceptions import DpaRequestTimeoutError, JsonRequestTimeoutError
from iqrfpy.messages import *
from iqrfpy.iresponse import IResponse
from iqrfpy.transports.mqtt_transport import MqttTransportParams, MqttTransport


def print_formatted(text, data):
    print(f'{text:30s} {data}')


def send_receive(request: IRequest, timeout: int) -> None:
    print('===========')
    print_formatted('Request sent at:', datetime.now())
    try:
        response = transport.send_and_receive(request=request, timeout=timeout)
        response_handler(response)
    except (DpaRequestTimeoutError, JsonRequestTimeoutError) as e:
        print_formatted('Error:', str(e))


def response_handler(response: IResponse) -> None:
    print_formatted('Received message at:', datetime.now())
    print_formatted('Message type:', response.mtype)
    print_formatted('Message ID:', response.msgid)
    print_formatted('Response code:', response.rcode)


params = MqttTransportParams(
        host='localhost',
        port=1883,
        client_id='python-lib-test',
        request_topic='Iqrf/DpaRequest',
        response_topic='Iqrf/DpaResponse',
        qos=1,
        keepalive=25
    )
transport = MqttTransport(params=params, auto_init=True)

time.sleep(5)

send_receive(LedrPulseReq(nadr=0, msgid='pulseTest'), timeout=2)

send_receive(OsReadReq(nadr=0, msgid='osTest1'), timeout=1)


print('===========')
print_formatted('Request sent at:', datetime.now())
transport.send(OsReadReq(nadr=0, msgid='osTest2'))
rsp = transport.receive(timeout=1)
response_handler(rsp)


try:
    print('===========')
    print_formatted('Request sent at:', datetime.now())
    rsp = transport.send_and_receive(OsReadReq(nadr=2, msgid='osTest3'), timeout=1)
    response_handler(rsp)
except (DpaRequestTimeoutError, JsonRequestTimeoutError) as error:
    print_formatted('Message not received:', str(error))



"""Generic response message."""

from typing import List, Optional, Union
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.enums.commands import Command
from iqrfpy.enums.message_types import GenericMessages
from iqrfpy.enums.peripherals import Peripheral
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.dpa import ResponsePacketMembers
from iqrfpy.utils.validators import DpaValidator, JsonValidator


class GenericResponse(IResponseGetterMixin):
    """Generic response class."""

    def __init__(self, nadr: int, pnum: Union[Peripheral, int], pcmd: Union[Command, int],
                 hwpid: int = dpa_constants.HWPID_MAX, rcode: int = 0, dpa_value: int = 0,
                 pdata: Optional[List[int]] = None, msgid: Optional[str] = None):
        """Generic response constructor.

        Args:
            nadr (int): Device address.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            msgid (str, optional): Message ID. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pnum=pnum,
            pcmd=pcmd,
            m_type=GenericMessages.RAW,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            msgid=msgid,
            pdata=pdata,
            result=pdata,
        )

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'GenericResponse':
        """DPA response factory method.

        Parses DPA data and constructs GenericResponse object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            GenericResponse: Response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        DpaValidator.response_length(dpa=dpa, expected_len=8)
        nadr = dpa[ResponsePacketMembers.NADR]
        pnum = dpa[ResponsePacketMembers.PNUM]
        pcmd = dpa[ResponsePacketMembers.PCMD]
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        pdata = Common.pdata_from_dpa(dpa)
        return cls(nadr=nadr, pnum=pnum, pcmd=pcmd, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata)

    @classmethod
    def from_json(cls, json: dict) -> 'GenericResponse':
        """JSON response factory method.

        Parses JSON API response and constructs GenericResponse object.

        Args:
            json (dict): JSON API response.
        Returns:
            GenericResponse: Response message object.
        """
        JsonValidator.response_received(json=json)
        msgid = Common.msgid_from_json(json=json)
        raw = [int(x, 16) for x in Common.generic_rdata_from_json(json=json).split('.')]
        nadr = raw[ResponsePacketMembers.NADR]
        pnum = raw[ResponsePacketMembers.PNUM]
        pcmd = raw[ResponsePacketMembers.PCMD]
        hwpid = Common.hwpid_from_dpa(raw[ResponsePacketMembers.HWPID_HI], raw[ResponsePacketMembers.HWPID_LO])
        rcode = raw[ResponsePacketMembers.RCODE]
        dpa_value = raw[ResponsePacketMembers.DPA_VALUE]
        pdata = raw[8:]
        return cls(nadr=nadr, pnum=pnum, pcmd=pcmd, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata,
                   msgid=msgid)



"""OS Read response message."""

from typing import List, Optional
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.enums.commands import OSResponseCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.os.responses.os_read_data import OsReadData
from iqrfpy.utils import dpa as dpa_constants
from iqrfpy.utils.dpa import ResponsePacketMembers, ResponseCodes
from iqrfpy.utils.common import Common
from iqrfpy.utils.validators import DpaValidator, JsonValidator

__all__ = ['ReadResponse']


class ReadResponse(IResponseGetterMixin):
    """OS Read response class."""

    __slots__ = ('_os_read_data',)

    def __init__(self, nadr: int, hwpid: int = dpa_constants.HWPID_MAX, rcode: int = 0, dpa_value: int = 0,
                 msgid: Optional[str] = None, pdata: Optional[List[int]] = None, result: Optional[dict] = None):
        """Read response constructor.

        Args:
            nadr (int): Device address.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            pdata (List[int], optional): DPA response data. Defaults to None.
            msgid (str, optional): Message ID. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.OS,
            pcmd=OSResponseCommands.READ,
            m_type=OSMessages.READ,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            msgid=msgid,
            pdata=pdata,
            result=result
        )
        self._os_read_data = OsReadData(data=result) if rcode == ResponseCodes.OK else None

    @property
    def os_read_data(self) -> Optional[OsReadData]:
        """:obj:`OsReadData` or :obj:`None`: OS Read data.

        Getter only.
        """
        return self._os_read_data

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'ReadResponse':
        """DPA response factory method.

        Parses DPA data and constructs :obj:`ReadResponse` object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            :obj:`ReadResponse`: Response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        nadr = dpa[ResponsePacketMembers.NADR]
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        pdata = None
        result = None
        if rcode == ResponseCodes.OK:
            if len(dpa) < 48:
                raise DpaResponsePacketLengthError('DPA response packet length invalid, expected at least 48B of data.')
            pdata = Common.pdata_from_dpa(dpa=dpa)
            result = {
                'mid': (dpa[11] << 24) + (dpa[10] << 16) + (dpa[9] << 8) + dpa[8],
                'osVersion': dpa[12],
                'trMcuType': dpa[13],
                'osBuild': (dpa[15] << 8) + dpa[14],
                'rssi': dpa[16],
                'supplyVoltage': 261.12 / (127 - dpa[17]),
                'flags': dpa[18],
                'slotLimits': dpa[19],
                'ibk': list(dpa[20:36]),
                'dpaVer': (dpa[37] << 8) + dpa[36],
                'perNr': dpa[38],
                'hwpid': (dpa[44] << 8) + dpa[43],
                'hwpidVer': (dpa[46] << 8) + dpa[45],
                'flagsEnum': dpa[47],
                'userPer': [],
            }
            embed_pers_data = list(dpa[39:43])
            embedded_pers = []
            for i in range(0, len(embed_pers_data * 8)):
                if embed_pers_data[int(i / 8)] & (1 << (i % 8)) and EmbedPeripherals.has_value(i):
                    embedded_pers.append(i)
            result['embeddedPers'] = embedded_pers
            if result['perNr'] > 0:
                user_per_data = list(dpa[48:])
                user_pers = []
                for i in range(0, len(user_per_data * 8)):
                    if user_per_data[int(i / 8)] & (1 << (i % 8)):
                        user_pers.append(i + 0x20)
                result['userPer'] = user_pers
        return cls(nadr=nadr, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata, result=result)

    @classmethod
    def from_json(cls, json: dict) -> 'ReadResponse':
        """JSON response factory method.

        Parses JSON API response and constructs :obj:`ReadResponse` object.

        Args:
            json (dict): JSON API response.
        Returns:
            :obj:`ReadResponse`: Response message object.
        """
        JsonValidator.response_received(json=json)
        nadr = Common.nadr_from_json(json=json)
        msgid = Common.msgid_from_json(json=json)
        hwpid = Common.hwpid_from_json(json=json)
        dpa_value = Common.dpa_value_from_json(json=json)
        rcode = Common.rcode_from_json(json=json)
        pdata = Common.pdata_from_json(json=json)
        result = Common.result_from_json(json=json) if rcode == ResponseCodes.OK else None
        return cls(nadr=nadr, msgid=msgid, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata, result=result)



"""FRC Set FRC Params request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import FrcRequestCommands
from iqrfpy.enums.message_types import FrcMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.irequest import IRequest
from iqrfpy.peripherals.frc.frc_params import FrcParams
import iqrfpy.utils.dpa as dpa_constants

__all__ = ['SetFrcParamsRequest']


class SetFrcParamsRequest(IRequest):
    """FRC Set Frc Params request class."""

    __slots__ = ('_frc_params',)

    def __init__(self, nadr: int, frc_params: Union[FrcParams, int], hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Send request constructor.

        Args:
            nadr (int): Device address.
            frc_params (Union[FrcParams, int]): FRC parameters.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(frc_params)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.FRC,
            pcmd=FrcRequestCommands.SET_PARAMS,
            m_type=FrcMessages.SET_PARAMS,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._frc_params = frc_params

    @staticmethod
    def _validate(frc_params: Union[FrcParams, int]):
        """Validate FRC params parameter.

        Args:
            frc_params (Union[FrcParams, int]): FRC parameters.
        Raises:
            RequestParameterInvalidValueError: If frc_params is less than 0 or greater than 255.
        """
        if isinstance(frc_params, int) and not dpa_constants.BYTE_MIN <= frc_params <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('FRC params value should be between 0 and 255.')

    @property
    def frc_params(self) -> FrcParams:
        """:obj:`FrcParams`: FRC parameters.

        Getter and setter.
        """
        return self._frc_params

    @frc_params.setter
    def frc_params(self, value: Union[FrcParams, int]):
        self._validate(value)
        self._frc_params = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._frc_params if isinstance(self._frc_params, int) else self._frc_params.to_data()]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {
            'frcResponseTime': self._frc_params if isinstance(self._frc_params, int) else self._frc_params.to_data()
        }
        return super().to_json()



import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.enums.commands import FrcResponseCommands
from iqrfpy.enums.message_types import FrcMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.frc.responses.extra_result import ExtraResultResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': FrcMessages.EXTRA_RESULT,
    'msgid': 'extraResultTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 64,
    'result': {
        'frcData': [0] * 9
    },
    'dpa': b'\x00\x00\x0d\x81\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00'
}

data_ok_1: dict = {
    'mtype': FrcMessages.EXTRA_RESULT,
    'msgid': 'extraResultTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'frcData': [17, 20, 0, 14, 1, 2, 7, 8, 255]
    },
    'dpa': b'\x00\x00\x0d\x81\x02\x04\x00\x23\x11\x14\x00\x0e\x01\x02\x07\x08\xff'
}

data_error: dict = {
    'mtype': FrcMessages.EXTRA_RESULT,
    'msgid': 'extraResultTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x00\x80\x04\x04\x01\x23'
}


class ExtraResultResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, ExtraResultResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, ExtraResultResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, ExtraResultResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, ExtraResultResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, ExtraResultResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, ExtraResultResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.FRC)
        with self.subTest():
            self.assertEqual(response.pcmd, FrcResponseCommands.EXTRA_RESULT)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, FrcMessages.EXTRA_RESULT)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            ExtraResultResponse.from_dpa(b'\x00\x00\x0d\x81\x00\x00\x00\x40\x0a')

    @parameterized.expand([
        ['from_dpa', data_ok['result']['frcData'], ExtraResultResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['frcData'], ExtraResultResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['frcData'], ExtraResultResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['frcData'], ExtraResultResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_frc_data(self, _, frc_data: List[int], response: ExtraResultResponse):
        self.assertEqual(response.frc_data, frc_data)

    @parameterized.expand([
        ['from_dpa_error', ExtraResultResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ExtraResultResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_frc_data_error(self, _, response: ExtraResultResponse):
        self.assertIsNone(response.frc_data)



r"""Enums module.

This module contains peripheral, command and message type enums.

Submodules
----------
:obj:`iqrfpy.enums.peripherals`: Peripherals module.\n
:obj:`iqrfpy.enums.commands`: Peripheral commands module.\n
:obj:`iqrfpy.enums.message_types`: Message types module.\n
"""



import unittest
from parameterized import parameterized
from iqrfpy.exceptions import JsonMsgidMissingError, JsonMTypeMissingError, UnsupportedMessageTypeError
from iqrfpy.response_factory import *
from iqrfpy.async_response import AsyncResponse
from iqrfpy.confirmation import Confirmation
from iqrfpy.peripherals.coordinator.responses import *


class ResponseFactoryTestCase(unittest.TestCase):

    @parameterized.expand([
        ['Async', b'\x00\x00\xff\x3f\x00\x00\x80\x00\x17\x04\x00\xfd\x26\x00\x00\x00\x00\x00\x00\x05', AsyncResponse],
        ['Confirmation', b'\x01\x00\x01\x00\xff\xff\xff\x36\x01\x04\x01', Confirmation],
        ['AddrInfo', b'\x00\x00\x00\x80\x00\x00\x00\x40\x0a\x2a', AddrInfoResponse],
        ['ClearAllBonds', b'\x00\x00\x00\x83\x00\x00\x00\x40', ClearAllBondsResponse],
        ['SmartConnect', b'\x00\x00\x00\x92\x00\x00\x00\x47\x01\x02', SmartConnectResponse]
    ])
    def test_response_from_dpa_ok(self, _, dpa, expected):
        self.assertIsInstance(ResponseFactory.get_response_from_dpa(dpa=dpa), expected)

    @parameterized.expand([
        ['Nonsense', b'\x0b\x0a']
    ])
    def test_response_from_dpa_invalid(self, _, dpa):
        with self.assertRaises(ValueError):
            ResponseFactory.get_response_from_dpa(dpa=dpa)

    @parameterized.expand([
        ['Async', {
            'mType': 'iqrfRaw',
            'data': {
                'msgId': 'async',
                'rsp': {
                    'rData': '00.00.ff.3f.00.00.80.00.17.04.00.fd.26.00.00.00.00.00.00.05'
                },
                'status': 0,
                'insId': 'iqrfgd2-default'
            }
        }, AsyncResponse],
        ['BondNode', {
            "mType": "iqrfEmbedCoordinator_BondNode",
            "data": {
                "msgId": "testEmbedCoordinator",
                "rsp": {
                    "nAdr": 0,
                    "hwpId": 0,
                    "rCode": 0,
                    "dpaVal": 53,
                    "result": {
                        "bondAddr": 1,
                        "devNr": 2
                    }
                },
                "insId": "iqrfgd2-1",
                "status": 0
            }
        }, BondNodeResponse],
        ['RemoveBond', {
            "mType": "iqrfEmbedCoordinator_RemoveBond",
            "data": {
                "msgId": "testEmbedCoordinator",
                "rsp": {
                    "nAdr": 0,
                    "hwpId": 0,
                    "rCode": 0,
                    "dpaVal": 53,
                    "result": {
                        "devNr": 1
                    }
                },
                "insId": "iqrfgd2-1",
                "status": 0
            }
        }, RemoveBondResponse],
        ['SetHops', {
            "mType": "iqrfEmbedCoordinator_SetHops",
            "data": {
                "msgId": "testEmbedCoordinator",
                "rsp": {
                    "nAdr": 0,
                    "hwpId": 0,
                    "rCode": 0,
                    "dpaVal": 0,
                    "result": {
                        "requestHops": 255,
                        "responseHops": 255
                    }
                },
                "insId": "iqrfgd2-1",
                "status": 0
            }
        }, SetHopsResponse]
    ])
    def test_response_from_json_ok(self, _, json, expected):
        self.assertIsInstance(ResponseFactory.get_response_from_json(json=json), expected)

    @parameterized.expand([
        [{'data': {'req': {}}}, JsonMsgidMissingError],
        [{'data': {'msgId': 'test'}}, JsonMTypeMissingError],
        [{'mType': 'unsupported', 'data': {'msgId': 'test'}}, UnsupportedMessageTypeError],
    ])
    def test_response_from_json_invalid(self, json, exception):
        with self.assertRaises(exception):
            ResponseFactory.get_response_from_json(json=json)




import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.bond_node import BondNodeRequest


class BondNodeRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x04\xff\xff\x01\x00'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_BondNode',
            'data': {
                'msgId': 'bondNodeTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'reqAddr': 1,
                        'bondingMask': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['req_addr', 1, 1, b'\x00\x00\x00\x04\xff\xff\x01\x01'],
        ['req_addr', 239, 1, b'\x00\x00\x00\x04\xff\xff\xef\x01'],
        ['bonding_test_retries', 1, 10, b'\x00\x00\x00\x04\xff\xff\x01\x0a'],
        ['bonding_test_retries', 1, 5, b'\x00\x00\x00\x04\xff\xff\x01\x05']
    ])
    def test_to_dpa(self, _, req_addr, bonding_test_retries, expected):
        request = BondNodeRequest(req_addr=req_addr, bonding_test_retries=bonding_test_retries)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['req_addr', 1, 1],
        ['req_addr', 239, 1],
        ['bonding_test_retries', 1, 10],
        ['bonding_test_retries', 1, 5],
        ['iquip', 240, 0]
    ])
    def test_to_json(self, _, req_addr: int, bonding_test_retries: int):
        request = BondNodeRequest(req_addr=req_addr, bonding_test_retries=bonding_test_retries, msgid='bondNodeTest')
        self.json['data']['req']['param']['reqAddr'] = req_addr
        self.json['data']['req']['param']['bondingMask'] = bonding_test_retries
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [239, b'\x00\x00\x00\x04\xff\xff\xef\x00'],
        [10, b'\x00\x00\x00\x04\xff\xff\x0a\x00']
    ])
    def test_set_req_addr(self, req_addr, dpa):
        request = BondNodeRequest(req_addr=1, bonding_test_retries=0, msgid='bondNodeTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.req_addr = req_addr
        self.json['data']['req']['param']['reqAddr'] = req_addr
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [1, b'\x00\x00\x00\x04\xff\xff\x01\x01'],
        [5, b'\x00\x00\x00\x04\xff\xff\x01\x05']
    ])
    def test_set_bonding_test_retries(self, bonding_test_retries, dpa):
        request = BondNodeRequest(req_addr=1, bonding_test_retries=0, msgid='bondNodeTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.bonding_test_retries = bonding_test_retries
        self.json['data']['req']['param']['bondingMask'] = bonding_test_retries
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, 1],
        [256, 1],
        [1000, 0],
        [1, -1],
        [1, 420]
    ])
    def test_construct_invalid(self, req_addr, bonding_test_retries):
        with self.assertRaises(ValueError):
            BondNodeRequest(req_addr=req_addr, bonding_test_retries=bonding_test_retries)



from iqrfpy.peripherals.coordinator.responses import *
from iqrfpy.response_factory import ResponseFactory


def handle_addr_info_response(response: AddrInfoResponse) -> None:
    print(f'peripheral: {response.pnum}')
    print(f'peripheral command: {response.pcmd}')
    status = response.rcode
    if status == 0:
        print(f'Addr info response dev_nr: {response.dev_nr}')
        print(f'Addr info response did: {response.did}')


def main():
    responses = [
        b'\x00\x00\x00\x80\x00\x00\x00\x40\x0a\x2a',
        b'\x00\x00\x00\x83\x00\x00\x00\x40',
        {
            "mType": "iqrfEmbedCoordinator_AddrInfo",
            "data": {
                "msgId": "testEmbedCoordinator",
                "rsp": {
                    "nAdr": 0,
                    "hwpId": 0,
                    "rCode": 0,
                    "dpaVal": 64,
                    "result": {
                        "devNr": 0,
                        "did": 42
                    }
                },
                "insId": "iqrfgd2-1",
                "status": 0
            }
        },
        {
            "mType": "iqrfEmbedCoordinator_RemoveBond",
            "data": {
                "msgId": "testEmbedCoordinator",
                "rsp": {
                    "nAdr": 0,
                    "hwpId": 0,
                    "rCode": 0,
                    "dpaVal": 53,
                    "result": {
                        "devNr": 1
                    }
                },
                "insId": "iqrfgd2-1",
                "status": 0
            }
        }
    ]
    for response in responses:
        rsp = None
        if isinstance(response, bytes):
            rsp = ResponseFactory.get_response_from_dpa(dpa=response)
        elif isinstance(response, dict):
            rsp = ResponseFactory.get_response_from_json(json=response)
        if isinstance(rsp, AddrInfoResponse):
            handle_addr_info_response(response=rsp)
        else:
            print(f'Unexpected response: {type(rsp)}')


if __name__ == '__main__':
    main()



"""Peripherals submodule containing specific peripherals and their messages."""

from . import generic
from . import binaryoutput
from . import coordinator
from . import eeeprom
from . import eeprom
from . import exploration
from . import frc
from . import io
from . import ledg
from . import ledr
from . import node
from . import os
from . import ram
from . import sensor
from . import thermometer
from . import uart



"""IO Direction request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import IORequestCommands
from iqrfpy.enums.message_types import IOMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest
from iqrfpy.peripherals.io.requests.io_triplet import IoTriplet

__all__ = [
    'DirectionRequest',
    'IoTriplet',
]


class DirectionRequest(IRequest):
    """IO direction request class."""

    __slots__ = ('_triplets',)

    def __init__(self, nadr: int, triplets: List[IoTriplet], hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Direction request constructor.

        Args:
            nadr (int): Device address.
            triplets (List[IoTriplet): List of port, mask and value subcommands triplets.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_triplets(triplets=triplets)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.IO,
            pcmd=IORequestCommands.DIRECTION,
            m_type=IOMessages.DIRECTION,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._triplets = triplets

    @staticmethod
    def _validate_triplets(triplets: List[IoTriplet]):
        """Validate IO subcommands triplets parameter.

        Args:
            triplets (List[IoTriplet): List of port, mask and value subcommands triplets.
        Raises:
            RequestParameterInvalidValueError: If list contains more than 18 triplets.
        """
        if len(triplets) > 18:
            raise RequestParameterInvalidValueError('Request can carry at most 18 triplets.')

    @property
    def triplets(self) -> List[IoTriplet]:
        """:obj:`list` of :obj:`IoTriplet`: List of port, mask and value subcommands triplets.

        Getter and setter.
        """
        return self._triplets

    @triplets.setter
    def triplets(self, val: List[IoTriplet]):
        self._validate_triplets(triplets=val)
        self._triplets = val

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        pdata = []
        for triplet in self._triplets:
            pdata.extend([triplet.port, triplet.mask, triplet.value])
        self._pdata = pdata
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        ports = [{'port': triplet.port, 'mask': triplet.mask, 'value': triplet.value} for triplet in self._triplets]
        self._params = {'ports': ports}
        return super().to_json()



import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.set_mid import SetMidRequest


class SetMidRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x13\xff\xff\x00\x00\x00\x00\x0a'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_SetMID',
            'data': {
                'msgId': 'setMIDTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'bondAddr': 10,
                        'mid': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['bond_addr', 5, 0, b'\x00\x00\x00\x13\xff\xff\x00\x00\x00\x00\x05'],
        ['bond_addr', 239, 0, b'\x00\x00\x00\x13\xff\xff\x00\x00\x00\x00\xef'],
        ['mid', 10, 2164554855, b'\x00\x00\x00\x13\xff\xff\x67\x7c\x04\x81\x0a'],
        ['mid', 10, 2164554771, b'\x00\x00\x00\x13\xff\xff\x13\x7c\x04\x81\x0a']
    ])
    def test_to_dpa(self, _, bond_addr, mid, expected):
        request = SetMidRequest(bond_addr=bond_addr, mid=mid)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['bond_addr', 5, 0],
        ['bond_addr', 239, 0],
        ['mid', 10, 2164554855],
        ['mid', 10, 2164554771]
    ])
    def test_to_json(self, _, bond_addr, mid):
        request = SetMidRequest(bond_addr=bond_addr, mid=mid, msgid='setMIDTest')
        self.json['data']['req']['param']['bondAddr'] = bond_addr
        self.json['data']['req']['param']['mid'] = mid
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [5, b'\x00\x00\x00\x13\xff\xff\x00\x00\x00\x00\x05'],
        [239, b'\x00\x00\x00\x13\xff\xff\x00\x00\x00\x00\xef']
    ])
    def test_set_bond_addr(self, bond_addr, dpa):
        request = SetMidRequest(bond_addr=10, mid=0, msgid='setMIDTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.bond_addr = bond_addr
        self.json['data']['req']['param']['bondAddr'] = bond_addr
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [2164554855, b'\x00\x00\x00\x13\xff\xff\x67\x7c\x04\x81\x0a'],
        [2164554771, b'\x00\x00\x00\x13\xff\xff\x13\x7c\x04\x81\x0a']
    ])
    def test_set_mid(self, mid, dpa):
        request = SetMidRequest(bond_addr=10, mid=0, msgid='setMIDTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.mid = mid
        self.json['data']['req']['param']['mid'] = mid
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, 2164554855],
        [257, 2164554855],
        [5, -1],
        [10, 4294967297]
    ])
    def test_construct_invalid(self, bond_addr, mid):
        with self.assertRaises(ValueError):
            SetMidRequest(bond_addr=bond_addr, mid=mid)



"""Sensor constants module.

This module provides constants related to Sensor standards.
"""

from enum import Enum
from iqrfpy.utils.enums import IntEnumMember


class SensorTypes(IntEnumMember):
    """Sensor type constants enum."""

    TEMPERATURE = 1
    CARBON_DIOXIDE = 2
    VOLATILE_ORGANIC_COMPOUND = 3
    EXTRA_LOW_VOLTAGE = 4
    EARTHS_MAGNETIC_FIELD = 5
    LOW_VOLTAGE = 6
    CURRENT = 7
    POWER = 8
    MAINS_FREQUENCY = 9
    TIME_SPAN = 10
    ILLUMINANCE = 11
    NITROGEN_DIOXIDE = 12
    SULFUR_DIOXIDE = 13
    CARBON_MONOXIDE = 14
    OZONE = 15
    ATMOSPHERIC_PRESSURE = 16
    COLOR_TEMPERATURE = 17
    PARTICULATES_PM2_5 = 18
    SOUND_PRESSURE_LEVEL = 19
    ALTITUDE = 20
    ACCELERATION = 21
    AMMONIA = 22
    METHANE = 23
    SHORT_LENGTH = 24
    PARTICULATES_PM1 = 25
    PARTICULATES_PM4 = 26
    PARTICULATES_PM10 = 27
    TOTAL_VOLATILE_ORGANIC_COMPOUND = 28
    NITROGEN_OXIDES = 29
    RELATIVE_HUMIDITY = 128
    BINARY_DATA7 = 129
    POWER_FACTOR = 130
    UV_INDEX = 131
    PH = 132
    RSSI = 133
    BINARY_DATA30 = 160
    CONSUMPTION = 161
    DATETIME = 162
    TIME_SPAN_LONG = 163
    LATITUDE = 164
    LONGITUDE = 165
    TEMPERATURE_FLOAT = 166
    LENGTH = 167
    DATA_BLOCK = 192


class SensorFrcCommands(IntEnumMember):
    """Sensor FRC commands enum."""

    FRC_2BITS = 0x10
    FRC_1BYTE = 0x90
    FRC_2BYTES = 0xE0
    FRC_4BYTES = 0xF9


class SensorDataSize(IntEnumMember):
    """Sensor data size by FRC command enum."""

    DATA_2BYTES_MIN = 1
    DATA_2BYTES_MAX = 127
    DATA_1BYTE_MIN = 128
    DATA_1BYTE_MAX = 159
    DATA_4BYTES_MIN = 160
    DATA_4BYTES_MAX = 191


class SensorFrcErrors(Enum):
    """Sensor FRC error codes enum."""

    NO_FRC_RESPONSE = "No FRC response"
    FRC_NOT_IMPLEMENTED = "FRC not implemented"
    SENSOR_ERROR_OR_OUT_OF_RANGE = "Sensor error or out of range"
    RESERVED = "Reserved"

    def __str__(self):
        """Return formatted error string message."""
        return self.value

    @classmethod
    def from_int(cls, value: int) -> 'SensorFrcErrors':
        """Convert an integer value to SensorFrcError object.

        Args:
            value (int): Integer value to convert
        Returns:
            :obj:`SensorFrcErrors`: SensorFrcErrors enum member corresponding to the error code.
        Raises:
            ValueError: If value is not a SensorFrcError code.
        """
        match value:
            case 0:
                return cls.NO_FRC_RESPONSE
            case 1:
                return cls.FRC_NOT_IMPLEMENTED
            case 2:
                return cls.SENSOR_ERROR_OR_OUT_OF_RANGE
            case 3:
                return cls.RESERVED
            case _:
                raise ValueError('Invalid Sensor FRC error code.')



"""Node Read Data parameters."""

from dataclasses import dataclass


@dataclass
class NodeReadData:
    """Node Read Data class."""

    __slots__ = 'ntw_addr', 'ntw_vrn', 'ntw_zin', 'ntw_did', 'ntw_pvrn', 'ntw_useraddr', 'ntw_id', 'ntw_vrnfnz', \
        'ntw_cfg', 'flags'

    def __init__(self, data: dict):
        """Node Read Data constructor.

        Args:
            data (dict): Node read result.
        """
        self.ntw_addr: int = data['ntwADDR']
        """Logical node address."""
        self.ntw_vrn: int = data['ntwVRN']
        """Virtual routing number."""
        self.ntw_zin: int = data['ntwZIN']
        """Zone index (zone number + 1)."""
        self.ntw_did: int = data['ntwDID']
        """Discovery ID."""
        self.ntw_pvrn: int = data['ntwPVRN']
        """Parent virtual routing number."""
        self.ntw_useraddr: int = data['ntwUSERADDRESS']
        """User address."""
        self.ntw_id: int = data['ntwID']
        """Network identification."""
        self.ntw_vrnfnz: int = data['ntwVRNFNZ']
        """Virtual routing number of first node in zone."""
        self.ntw_cfg: int = data['ntwCFG']
        """Network configuration."""
        self.flags: int = data['flags']
        """Node flags."""



"""Coordinator Bond Node request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import CoordinatorRequestCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['BondNodeRequest']


class BondNodeRequest(IRequest):
    """Coordinator BondNode request class."""

    __slots__ = '_req_addr', '_bonding_test_retries'

    def __init__(self, req_addr: int, bonding_test_retries: int, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """BondNode request constructor.

        Args:
            req_addr (int): Requested node address.
            bonding_test_retries (int): Maximum number of FRCs used to test whether node was bonded.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(req_addr, bonding_test_retries)
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorRequestCommands.BOND_NODE,
            m_type=CoordinatorMessages.BOND_NODE,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._req_addr = req_addr
        self._bonding_test_retries = bonding_test_retries

    def _validate(self, req_addr: int, bonding_test_retries: int) -> None:
        """Validate request parameters.

        Addr:
            req_addr (int): Requested node address.
            bonding_test_retries (int): Maximum number of FRCs used to test whether node was bonded.
        """
        self._validate_req_addr(req_addr)
        self._validate_bonding_test_retries(bonding_test_retries)

    @staticmethod
    def _validate_req_addr(req_addr: int):
        """Validates requested address parameter.

        Args:
            req_addr (int): Requested node address.
        Raises:
            RequestParameterInvalidValueError: If req_addr is less than 0 or greater than 255.
        """
        if req_addr < dpa_constants.BYTE_MIN or req_addr > dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Address value should be between 0 and 255.')

    @property
    def req_addr(self) -> int:
        """:obj:`int`: Requested node address.

        Getter and setter.
        """
        return self._req_addr

    @req_addr.setter
    def req_addr(self, value: int):
        self._validate_req_addr(value)
        self._req_addr = value

    @staticmethod
    def _validate_bonding_test_retries(bonding_test_retries: int):
        """Validates bonding test retries parameter

        Args:
            bonding_test_retries (int): Maximum number of FRCs used to test whether node was bonded.
        Raises:
            RequestParameterInvalidValueError: If bonding_test_retries is less than 0 or greater than 255.
        """
        if bonding_test_retries < dpa_constants.BYTE_MIN or bonding_test_retries > dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Bonding test retries value should be between 0 and 255.')

    @property
    def bonding_test_retries(self) -> int:
        """:obj:`int`: Maximum number of FRCs used to test whether node was bonded.

        Getter and setter.
        """
        return self._bonding_test_retries

    @bonding_test_retries.setter
    def bonding_test_retries(self, value: int):
        self._validate_bonding_test_retries(value)
        self._bonding_test_retries = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._req_addr, self._bonding_test_retries]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'reqAddr': self._req_addr, 'bondingMask': self._bonding_test_retries}
        return super().to_json()



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import EEEPROMResponseCommands
from iqrfpy.enums.message_types import EEEPROMMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.eeeprom.responses import ReadResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': EEEPROMMessages.READ.value,
    'msgid': 'readTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 90,
    'result': {
        'pData': [10, 20, 30, 40, 1, 12]
    },
    'dpa': b'\x00\x00\x04\x82\x00\x00\x00\x5a\x0a\x14\x1e\x28\x01\x0c'
}

data_ok_1: dict = {
    'mtype': EEEPROMMessages.READ.value,
    'msgid': 'readTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'pData': [10, 20, 30, 40, 1, 12]
    },
    'dpa': b'\x00\x00\x04\x82\x02\x04\x00\x5a\x0a\x14\x1e\x28\x01\x0c'
}

data_error: dict = {
    'mtype': EEEPROMMessages.READ.value,
    'msgid': 'readTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 4,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x04\x82\x04\x04\x04\x23'
}


class ReadResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, ReadResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, ReadResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, ReadResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, ReadResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, ReadResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, ReadResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.EEEPROM)
        with self.subTest():
            self.assertEqual(response.pcmd, EEEPROMResponseCommands.READ)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, EEEPROMMessages.READ)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    @parameterized.expand([
        ['from_dpa', data_ok['result']['pData'], ReadResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['pData'], ReadResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['pData'], ReadResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['pData'], ReadResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_data(self, _, data, response: ReadResponse):
        self.assertEqual(response.data, data)

    @parameterized.expand([
        ['from_dpa_error', ReadResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_data_error(self, _, response: ReadResponse):
        self.assertIsNone(response.data)




"""OS Batch request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import OSRequestCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.os.requests.batch_data import BatchData
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['BatchRequest']


class BatchRequest(IRequest):
    """OS Batch request class."""

    __slots__ = ('_requests',)

    def __init__(self, nadr: int, requests: List[BatchData], hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Batch request constructor.

        Args:
            nadr (int): Device address.
            requests (List[BatchData]): Batch request data.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(requests)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.OS,
            pcmd=OSRequestCommands.SET_SECURITY,
            m_type=OSMessages.SET_SECURITY,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._requests = requests

    @staticmethod
    def _validate(requests: List[BatchData]):
        """Validate batch requests.

        Args:
            requests (List[BatchData]): Batch request data.
        """
        data = []
        for request in requests:
            data += request.to_pdata()
        if len(data) + 1 > 58:
            raise RequestParameterInvalidValueError('Batch requests data should be no larger than 58B.')

    @property
    def requests(self) -> List[BatchData]:
        """:obj:`list` of :obj:`BatchData`: Batch request data.

        Getter and setter.
        """
        return self._requests

    @requests.setter
    def requests(self, value: List[BatchData]):
        self._validate(value)
        self._requests = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = []
        for request in self._requests:
            self._pdata += request.to_pdata()
        self._pdata.append(0)
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'requests': [request.to_json() for request in self._requests]}
        return super().to_json()



import unittest
from typing import Union
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.frc.requests.set_frc_params import SetFrcParamsRequest, FrcParams
from iqrfpy.utils.dpa import FrcResponseTimes


class SetFrcParamsRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x0d\x03\xff\xff\x00'
        self.json = {
            'mType': 'iqrfEmbedFrc_SetParams',
            'data': {
                'msgId': 'setFrcParamsTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'frcResponseTime': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['int_params', 24, b'\x00\x00\x0d\x03\xff\xff\x18'],
        ['class_params',
         FrcParams(offline_frc=True, frc_response_time=FrcResponseTimes.MS360),
         b'\x00\x00\x0d\x03\xff\xff\x18'
         ]
    ])
    def test_to_dpa(self, _, params: Union[FrcParams, int], expected):
        request = SetFrcParamsRequest(nadr=0, frc_params=params)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['int_params', 24],
        ['class_params', FrcParams(offline_frc=True, frc_response_time=FrcResponseTimes.MS360)]
    ])
    def test_to_json(self, _, params: Union[FrcParams, int]):
        request = SetFrcParamsRequest(nadr=0, frc_params=params, msgid='setFrcParamsTest')
        self.json['data']['req']['param'] = {'frcResponseTime': params if type(params) == int else params.to_data()}
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        ['int_params', 24, b'\x00\x00\x0d\x03\xff\xff\x18'],
        ['class_params',
         FrcParams(offline_frc=True, frc_response_time=FrcResponseTimes.MS360),
         b'\x00\x00\x0d\x03\xff\xff\x18'
         ]
    ])
    def test_set_frc_params(self, _, params: Union[FrcParams, int], dpa):
        default_params = FrcParams(offline_frc=False, frc_response_time=FrcResponseTimes.MS40)
        request = SetFrcParamsRequest(nadr=0, frc_params=default_params, msgid='setFrcParamsTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.frc_params = params
        self.json['data']['req']['param'] = {'frcResponseTime': params if type(params) == int else params.to_data()}
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_params(self, params: int):
        with self.assertRaises(RequestParameterInvalidValueError):
            SetFrcParamsRequest(nadr=0, frc_params=params)



import random
import unittest
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.node.requests.restore import RestoreRequest


class RestoreRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x01\x07\xff\xff\x8d\x6c\x51\xab\x6a\xf0\xf7\xc8\x3d\x50\x36\x5e\x26\xe1\x3d\x5a\x46\x83' \
                   b'\x20\xbc\xd5\x23\x62\x74\x00\xf5\x80\xe2\xb4\x63\xa6\xad\xdf\xb8\x5a\x79\xad\xf4\x12\x5b\x55\x06' \
                   b'\xa9\x17\xea\x1b\xe1\x29\x06'
        self.json = {
            'mType': 'iqrfEmbedNode_Restore',
            'data': {
                'msgId': 'restoreTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'backupData': [
                            141,
                            108,
                            81,
                            171,
                            106,
                            240,
                            247,
                            200,
                            61,
                            80,
                            54,
                            94,
                            38,
                            225,
                            61,
                            90,
                            70,
                            131,
                            32,
                            188,
                            213,
                            35,
                            98,
                            116,
                            0,
                            245,
                            128,
                            226,
                            180,
                            99,
                            166,
                            173,
                            223,
                            184,
                            90,
                            121,
                            173,
                            244,
                            18,
                            91,
                            85,
                            6,
                            169,
                            23,
                            234,
                            27,
                            225,
                            41,
                            6
                        ]
                    }
                },
                'returnVerbose': True
            }
        }

    def test_to_dpa(self):
        backup_data = random.sample(range(0, 255), 49)
        expected = bytes(b'\x01\x00\x01\x07\xff\xff' + bytes(backup_data))
        request = RestoreRequest(nadr=1, backup_data=backup_data)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    def test_to_json(self):
        backup_data = random.sample(range(0, 255), 49)
        request = RestoreRequest(nadr=1, backup_data=backup_data, msgid='restoreTest')
        self.json['data']['req']['param']['backupData'] = backup_data
        self.assertEqual(
            request.to_json(),
            self.json
        )

    def test_set_backup_data(self):
        request = RestoreRequest(nadr=1, backup_data=self.json['data']['req']['param']['backupData'], msgid='restoreTest')
        print(request.to_dpa())
        print(self.dpa)
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        backup_data = random.sample(range(0, 255), 49)
        dpa = bytes(b'\x01\x00\x01\x07\xff\xff' + bytes(backup_data))
        request.backup_data = backup_data
        self.json['data']['req']['param']['backupData'] = backup_data
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[0] * 60],
        [[-255]],
        [[256]]
    ])
    def test_construct_invalid(self, backup_data):
        with self.assertRaises(RequestParameterInvalidValueError):
            RestoreRequest(nadr=1, backup_data=backup_data)



"""Common utility module.

common module provides utilities and auxiliary methods for
extraction of data from DPA bytes and Daemon API JSON messages.
"""

__all__ = ['Common']

import math
from typing import List, Union
from typeguard import typechecked
from iqrfpy.enums.commands import *
from iqrfpy.enums.message_types import *
from iqrfpy.enums.peripherals import *
from iqrfpy.exceptions import InvalidPeripheralValueError, InvalidPeripheralCommandValueError, \
    JsonMsgidMissingError, JsonMTypeMissingError, JsonNadrMissingError, JsonHwpidMissingError, JsonRCodeMissingError, \
    JsonDpaValueMissingError, JsonResultMissingError, JsonStatusMissingError, JsonGenericDataMissingError, \
    UnsupportedMessageTypeError, UnsupportedPeripheralError, UnsupportedPeripheralCommandError
import iqrfpy.utils.dpa as dpa_constants


@typechecked
class Common:
    """Common class provides auxiliary methods for handling DPA and Daemon API JSON messages."""

    # DPA

    @staticmethod
    def hwpid_from_dpa(high: int, low: int) -> int:
        """Convert DPA HWPID bytes to a single 16bit unsigned integer.

        Args:
            high (int): HWPID high byte
            low (int): HWPID low byte
        Returns:
            :obj:`int`: 16bit unsigned integer HWPID value
        Raises:
            ValueError: Raised if input values are not between 0 and 255
        """
        if high > dpa_constants.BYTE_MAX or low > dpa_constants.BYTE_MAX:
            raise ValueError('Argument value exceeds maximum allowed value of 255.')
        if high < dpa_constants.BYTE_MIN or low < dpa_constants.BYTE_MIN:
            raise ValueError('Negative argument values are not allowed.')
        return (high << 8) + low

    @staticmethod
    def pnum_from_dpa(pnum: int) -> Peripheral:
        """Return peripheral enum value based on DPA peripheral data byte.

        Args:
            pnum (int): Peripheral number data byte
        Returns:
            :obj:`Peripheral`: Peripheral enum member
        Raises:
            InvalidPeripheralValueError: Raised if pnum value is not between 0 and 255
            UnsupportedPeripheralError: Raised if pnum parameter value is not recognized as a member of any peripheral enum
        """
        if pnum < 0 or pnum > 255:
            raise InvalidPeripheralValueError('Peripheral value out of range 0-255.')
        if EmbedPeripherals.has_value(pnum):
            return EmbedPeripherals(pnum)
        if Standards.has_value(pnum):
            return Standards(pnum)
        raise UnsupportedPeripheralError('Unknown or unsupported peripheral.')

    @staticmethod
    def request_pcmd_from_dpa(pnum: Peripheral, pcmd: int) -> Command:
        """Return request command based on DPA peripheral and command data byte.

        Args:
            pnum (Peripheral): Peripheral enum member
            pcmd (int): Command data byte value
        Returns:
            :obj:`Command`: Request command enum value
        Raises:
            InvalidPeripheralCommandValueError: Raised if pcmd is a negative value, or if pcmd is not a value between 0 and 127
            UnsupportedPeripheralError: Raised if pnum parameter value is not recognized as a member of any peripheral enum
            UnsupportedPeripheralCommandError: Raised if pcmd parameter value is not recognized as a member of any peripheral command enum
        """
        if pcmd < dpa_constants.REQUEST_PCMD_MIN:
            raise InvalidPeripheralCommandValueError('Negative peripheral command values are not allowed.')
        if pcmd > dpa_constants.REQUEST_PCMD_MAX:
            raise InvalidPeripheralCommandValueError('Peripheral command value exceeds maximum allowed value of 127.')
        commands = None
        match pnum:
            case EmbedPeripherals.COORDINATOR:
                commands = CoordinatorRequestCommands
            case EmbedPeripherals.NODE:
                commands = NodeRequestCommands
            case EmbedPeripherals.OS:
                commands = OSRequestCommands
            case EmbedPeripherals.EEPROM:
                commands = EEPROMRequestCommands
            case EmbedPeripherals.EEEPROM:
                commands = EEEPROMRequestCommands
            case EmbedPeripherals.RAM:
                commands = RAMRequestCommands
            case EmbedPeripherals.LEDR | EmbedPeripherals.LEDG:
                commands = LEDRequestCommands
            case EmbedPeripherals.IO:
                commands = IORequestCommands
            case EmbedPeripherals.THERMOMETER:
                commands = ThermometerRequestCommands
            case EmbedPeripherals.UART:
                commands = UartRequestCommands
            case EmbedPeripherals.FRC:
                commands = FrcRequestCommands
            case EmbedPeripherals.EXPLORATION:
                commands = ExplorationRequestCommands
            case Standards.DALI:
                commands = DALIRequestCommands
            case Standards.BINARY_OUTPUT:
                commands = BinaryOutputRequestCommands
            case Standards.SENSOR:
                commands = SensorRequestCommands
            case Standards.LIGHT:
                commands = LightRequestCommands
            case _:
                raise UnsupportedPeripheralError('Unknown or unsupported peripheral.')

        if commands is not None and commands.has_value(pcmd):
            return commands(pcmd)
        raise UnsupportedPeripheralCommandError('Unknown or unsupported peripheral command.')

    @staticmethod
    def response_pcmd_from_dpa(pnum: Peripheral, pcmd: int) -> Command:
        """Return response command based on DPA peripheral and command data byte.

        Args:
            pnum (Peripheral): Peripheral enum member
            pcmd (int): Command data byte value
        Returns:
            :obj:`Command`: Response command enum member
        Raises:
            InvalidPeripheralCommandValueError: Raised if pcmd is a negative value, or if pcmd is not a value between 128 and 255
            UnsupportedPeripheralError: Raised if pnum parameter value is not recognized as a member of any peripheral enum
            UnsupportedPeripheralCommandError: Raised if pcmd parameter value is not recognized as a member of any peripheral command enum
        """
        if pcmd < dpa_constants.REQUEST_PCMD_MIN:
            raise InvalidPeripheralCommandValueError('Negative peripheral command values are not allowed.')
        if pcmd <= dpa_constants.REQUEST_PCMD_MAX or pcmd > dpa_constants.RESPONSE_PCMD_MAX:
            raise InvalidPeripheralCommandValueError('Response peripheral command should be value between 128 and 255.')
        commands = None
        match pnum:
            case EmbedPeripherals.COORDINATOR:
                commands = CoordinatorResponseCommands
            case EmbedPeripherals.NODE:
                commands = NodeResponseCommands
            case EmbedPeripherals.OS:
                commands = OSResponseCommands
            case EmbedPeripherals.EEPROM:
                commands = EEPROMResponseCommands
            case EmbedPeripherals.EEEPROM:
                commands = EEEPROMResponseCommands
            case EmbedPeripherals.RAM:
                commands = RAMResponseCommands
            case EmbedPeripherals.LEDR | EmbedPeripherals.LEDG:
                commands = LEDResponseCommands
            case EmbedPeripherals.IO:
                commands = IOResponseCommands
            case EmbedPeripherals.THERMOMETER:
                commands = ThermometerResponseCommands
            case EmbedPeripherals.UART:
                commands = UartResponseCommands
            case EmbedPeripherals.FRC:
                commands = FrcResponseCommands
            case EmbedPeripherals.EXPLORATION:
                commands = ExplorationResponseCommands
            case Standards.DALI:
                commands = DALIResponseCommands
            case Standards.BINARY_OUTPUT:
                commands = BinaryOutputResponseCommands
            case Standards.SENSOR:
                commands = SensorResponseCommands
            case Standards.LIGHT:
                commands = LightResponseCommands
            case _:
                raise UnsupportedPeripheralError('Unknown or unsupported peripheral.')

        if commands is not None and commands.has_value(pcmd):
            return commands(pcmd)
        raise UnsupportedPeripheralCommandError('Unknown or unsupported peripheral command.')

    @staticmethod
    def pdata_from_dpa(dpa: bytes) -> Union[List[int], None]:
        """Return PDATA from DPA response bytes.

        Args:
            dpa (bytes): DPA response message
        Returns:
            :obj:`list` of :obj:`int` or :obj:`None`: PDATA integer list or None of there are no PDATA
        """
        if len(dpa) > 8:
            return list(dpa[8:])
        return None

    # json

    @staticmethod
    def msgid_from_json(json: dict) -> str:
        """Return response msgid from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`str`: JSON API response message ID
        Raises:
            JsonMsgidMissingError: Raised if Daemon API response does not contain the msgId key
        """
        try:
            return json['data']['msgId']
        except KeyError as err:
            raise JsonMsgidMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def mtype_str_from_json(json: dict) -> str:
        """Return message type from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`str`: JSON API response message type string
        Raises:
            JsonMTypeMissingError: Raised if Daemon API response does not contain the mType key
        """
        try:
            return json['mType']
        except KeyError as err:
            raise JsonMTypeMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def nadr_from_json(json: dict) -> int:
        """Return response nadr from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response device address
        Raises:
            JsonNadrMissingError: Raised if Daemon API response does not contain the nAdr key
        """
        try:
            return json['data']['rsp']['nAdr']
        except KeyError as err:
            raise JsonNadrMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def pnum_from_json(json: dict) -> int:
        """Return response pnum from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response peripheral number
        Raises:
            JsonNadrMissingError: Raised if Daemon API response does not contain the pnum key
        """
        try:
            return json['data']['rsp']['pnum']
        except KeyError as err:
            raise JsonNadrMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def pcmd_from_json(json: dict) -> int:
        """Return response pcmd from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response peripheral command
        Raises:
            JsonNadrMissingError: Raised if Daemon API response does not contain the pcmd key
        """
        try:
            return json['data']['rsp']['pcmd']
        except KeyError as err:
            raise JsonNadrMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def hwpid_from_json(json: dict) -> int:
        """Return response hwpid from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response hardware profile ID
        Raises:
            JsonHwpidMissingError: Raised if Daemon API response does not contain the hwpId key
        """
        try:
            return json['data']['rsp']['hwpId']
        except KeyError as err:
            raise JsonHwpidMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def rcode_from_json(json: dict) -> int:
        """Return response rcode from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response DPA rcode
        Raises:
            JsonRCodeMissingError: Raised if Daemon API response does not contain the rcode key
        """
        try:
            return json['data']['rsp']['rCode']
        except KeyError as err:
            raise JsonRCodeMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def dpa_value_from_json(json: dict) -> int:
        """Return response DPA value from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response dpa value
        Raises:
            JsonDpaValueMissingError: Raised if Daemon API response does not contain the dpaVal key
        """
        try:
            return json['data']['rsp']['dpaVal']
        except KeyError as err:
            raise JsonDpaValueMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def pdata_from_json(json: dict) -> Union[List[int], None]:
        """Return pdata from Daemon API JSON response if available.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`Union[List[int], None]`: JSON API response pdata or None if there are no PDATA
        """
        pdata = None
        try:
            raw = json['data']['raw']
            response = None
            if isinstance(raw, list):
                response = raw[0]['response']
            elif isinstance(raw, dict):
                response = raw['response']
            response = response.split('.')
            if len(response) > 8:
                pdata = [int(x, 16) for x in response[8:]]
        except KeyError:
            pdata = None
        finally:
            return pdata

    @staticmethod
    def result_from_json(json: dict) -> dict:
        """Return response result from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`dict`: JSON API response result object
        Raises:
            JsonResultMissingError: Raised if JSON API response does not contain the result key
        """
        try:
            return json['data']['rsp']['result']
        except KeyError as err:
            raise JsonResultMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def status_from_json(json: dict) -> int:
        """Return response status from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response status code
        Raises:
            JsonStatusMissingError: Raised if JSON API response does not contain the status key
        """
        try:
            return json['data']['status']
        except KeyError as err:
            raise JsonStatusMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def generic_rdata_from_json(json: dict) -> str:
        """Return generic response data from Daemon API JSON response.

        Args:
            json (dict): JSON API response
        Returns:
            :obj:`int`: JSON API response status code
        Raises:
            JsonStatusMissingError: Raised if JSON API response does not contain path to rData.
        """
        try:
            return json['data']['rsp']['rData']
        except KeyError as err:
            raise JsonGenericDataMissingError(f'Object does not contain property {str(err)}') from err

    @staticmethod
    def string_to_mtype(string: str) -> MessageType:
        """Convert message type string to message type enum member value.

        Args:
            string (str): Message type string
        Returns:
            :obj:`MessageType`: Message type enum member
        Raises:
            UnsupportedMessageTypeError: Raised if message type is not recognized as a member of any message type enum
        """
        messages = [GenericMessages, ExplorationMessages, CoordinatorMessages, NodeMessages, OSMessages, EEPROMMessages,
                    EEEPROMMessages, RAMMessages, LEDRMessages, LEDGMessages, IOMessages, ThermometerMessages,
                    UartMessages, FrcMessages, DALIMessages, BinaryOutputMessages, SensorMessages, LightMessages]
        for item in messages:
            if item.has_value(string):
                return item(string)
        raise UnsupportedMessageTypeError(f'Unknown or unsupported message type: {string}.')

    # general

    @staticmethod
    def bitmap_to_nodes(bitmap: List[int], coordinator_shift: bool = False) -> List[int]:
        """Convert node bitmap to list of nodes.

        Args:
            bitmap (List[int]): Node bitmap represented by list of integers
            coordinator_shift (bool): Bitmap contains dummy coordinator value
        Returns:
            :obj:`list` of :obj:`int`: List of node addresses from bitmap
        """
        nodes = []
        start = 0 if not coordinator_shift else 1
        for i in range(start, len(bitmap * 8)):
            if bitmap[int(i / 8)] & (1 << (i % 8)):
                nodes.append(i)
        return nodes

    @staticmethod
    def nodes_to_bitmap(nodes: List[int]) -> List[int]:
        """Convert list of nodes to node bitmap.

        Args:
            nodes (List[int]): List of node addresses
        Returns:
            :obj:`list` of :obj:`int`: Nodes bitmap represented by list of 30 integers
        """
        bitmap = [0] * 30
        for node in nodes:
            bitmap[math.floor(node / 8)] |= (1 << (node % 8))
        return bitmap

    @staticmethod
    def bitmap_4byte_to_indexes(bitmap: List[int]) -> List[int]:
        """Convert 4 byte bitmap to list of indexes.

        Args:
            bitmap (List[int]): Index bitmap.
        Returns:
            :obj:`List` of :obj:`int`: List of indexes.
        """
        indexes = [0] * 32
        for i in range(0, len(bitmap * 8)):
            if bitmap[int(i / 8)] & (1 << (i % 8)):
                indexes[i] = 1
        return indexes

    @staticmethod
    def indexes_to_4byte_bitmap(values: List[int]) -> List[int]:
        """Convert list of indexes to a 4 byte bitmap.

        Args:
            values (List[int]): List of indexes
        Returns:
            :obj:`list` of :obj:`int`: Index bitmap represented by list of 4 integers
        """
        bitmap = [0] * 4
        for value in values:
            bitmap[math.floor(value / 8)] |= (1 << (value % 8))
        return bitmap

    @staticmethod
    def is_hex_string(string: str) -> bool:
        """Check if string contains only hexadecimal characters.

        Args:
            string (str): Input string
        Returns:
            :obj:`bool`: True if string contains only hexadecimal characters, False otherwise
        """
        if len(string) == 0:
            return False
        return not set(string) - set('0123456789abcdefABCDEF')

    @staticmethod
    def hex_string_to_list(string: str) -> List[int]:
        """Convert hexadecimal string to list of unsigned integers.

        Args:
            string (str): Hexadecimal string
        Returns:
            :obj:`list` of :obj:`int`: List of integers from hexadecimal string
        Raises:
            ValueError: Raised if string is of uneven length or contains non-hexadecimal characters
        """
        if not len(string) % 2 == 0:
            raise ValueError('Argument should be even length.')
        if not Common.is_hex_string(string):
            raise ValueError('Argument is not a hexadecimal string.')
        return [int(string[i:i + 2], base=16) for i in range(0, len(string), 2)]

    @staticmethod
    def list_to_hex_string(values: List[int]) -> str:
        """Convert list of unsigned integers to hexadecimal string.

        Args:
            values (List[int]): List of unsigned integers
        Returns:
            :obj:`string`: Hexadecimal string representation of the list
        """
        return ''.join(f'{value:02x}' for value in values)

    @staticmethod
    def peripheral_list_to_bitmap(values: List[int]) -> List[int]:
        """Convert list of peripheral numbers into a bitmap of peripherals.

        For example, the coordinator peripheral would be represented by the 1st bit in bitmap.

        Args:
            values (List[int]): List of peripheral numbers
        Returns:
            :obj:`list` of :obj:`int`: Peripheral bitmap represented by list of 4 integers
        """
        bitmap = [0 for _ in range(32)]
        for value in values:
            bitmap[value] = 1
        byte_list = []
        for bits in [bitmap[i:i + 8] for i in range(0, len(bitmap), 8)]:
            bits.reverse()
            byte = 0
            for bit in bits:
                byte = (byte << 1) | bit
            byte_list.append(byte)
        return byte_list

    @staticmethod
    def values_in_byte_range(values: List[int]) -> bool:
        """Check if list elements are within unsigned integer byte range.

        Args:
            values (List[int]): Input data
        Returns:
            :obj:`bool`: True if values are in range, False otherwise
        """
        return len([value for value in values if value < 0 or value > 255]) == 0

    @staticmethod
    def byte_complement(value: int):
        """Convert unsigned 1B value into a signed 1B value.

        Args:
            value (int): Input unsigned 1B value
        Returns:
            :obj:`int`: Signed 1B value
        """
        if not 0 <= value <= 0xFF:
            raise ValueError('Not an unsigned 1B value.')
        if value < 0x80:
            return value
        return value - 0x100

    @staticmethod
    def word_complement(value: int) -> int:
        """Convert unsigned 2B value into a signed 2B value.

        Args:
            value (int): Input unsigned 2B value
        Returns:
            :obj:`int`: Signed 2B value
        """
        if not 0 <= value <= 0xFFFF:
            raise ValueError('Not an unsigned 2B value.')
        if value < 0x8000:
            return value
        return value - 0x10000



"""Coordinator Set Hops request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import CoordinatorRequestCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['SetHopsRequest']


class SetHopsRequest(IRequest):
    """Coordinator Set Hops request class."""

    __slots__ = '_request_hops', '_response_hops'

    def __init__(self, request_hops: int, response_hops: int, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Set Hops request constructor.

        Args:
            request_hops (int): Number of hops to send DPA request.
            response_hops (int): Number of hops to send DPA response.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(request_hops, response_hops)
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorRequestCommands.SET_HOPS,
            m_type=CoordinatorMessages.SET_HOPS,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._request_hops = request_hops
        self._response_hops = response_hops

    def _validate(self, request_hops: int, response_hops: int) -> None:
        """Validates request parameters.

        Args:
            request_hops (int): Number of hops to send DPA request.
            response_hops (int): Number of hops to send DPA response.
        """
        self._validate_request_hops(request_hops)
        self._validate_response_hops(response_hops)

    @staticmethod
    def _validate_request_hops(request_hops: int):
        """Validates request hops parameter.

        Args:
            request_hops (int): Number of hops to send DPA request.
        Raises:
            RequestParameterInvalidValueError: If request_hops is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= request_hops <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Request hops value should be between 0 and 255.')

    @property
    def request_hops(self) -> int:
        """:obj:`int`: Number of hops to send DPA request.

        Getter and setter.
        """
        return self._request_hops

    @request_hops.setter
    def request_hops(self, value: int) -> None:
        self._validate_request_hops(value)
        self._request_hops = value

    @staticmethod
    def _validate_response_hops(response_hops: int):
        """Validates response hops parameter.

        Args:
            response_hops (int): Number of hops to send DPA response.
        Raises:
            RequestParameterInvalidValueError: If response_hops is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= response_hops <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Response hops value should be between 0 and 255.')

    @property
    def response_hops(self) -> int:
        """:obj:`int`: Number of hops to send DPA response.

        Getter and setter.
        """
        return self._response_hops

    @response_hops.setter
    def response_hops(self, value: int) -> None:
        self._validate_response_hops(value)
        self._response_hops = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._request_hops, self._response_hops]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'requestHops': self._request_hops, 'responseHops': self._response_hops}
        return super().to_json()



import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.set_hops import SetHopsRequest


class SetHopsRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x09\xff\xff\xff\xff'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_SetHops',
            'data': {
                'msgId': 'setHopsTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'requestHops': 255,
                        'responseHops': 255
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['requestHops', 255, 255, b'\x00\x00\x00\x09\xff\xff\xff\xff'],
        ['requestHops', 34, 255, b'\x00\x00\x00\x09\xff\xff\x22\xff'],
        ['responseHops', 255, 10, b'\x00\x00\x00\x09\xff\xff\xff\x0a'],
        ['responseHops', 255, 239, b'\x00\x00\x00\x09\xff\xff\xff\xef']
    ])
    def test_to_dpa(self, _, request_hops, response_hops, expected):
        request = SetHopsRequest(request_hops=request_hops, response_hops=response_hops)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['requestHops', 255, 255],
        ['requestHops', 34, 255],
        ['responseHops', 255, 10],
        ['responseHops', 255, 239]
    ])
    def test_to_json(self, _, request_hops, response_hops):
        request = SetHopsRequest(request_hops=request_hops, response_hops=response_hops, msgid='setHopsTest')
        self.json['data']['req']['param']['requestHops'] = request_hops
        self.json['data']['req']['param']['responseHops'] = response_hops
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [1, b'\x00\x00\x00\x09\xff\xff\x01\xff'],
        [239, b'\x00\x00\x00\x09\xff\xff\xef\xff']
    ])
    def test_set_request_hops(self, request_hops, dpa):
        request = SetHopsRequest(request_hops=255, response_hops=255, msgid='setHopsTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.request_hops = request_hops
        self.json['data']['req']['param']['requestHops'] = request_hops
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [10, b'\x00\x00\x00\x09\xff\xff\xff\x0a'],
        [239, b'\x00\x00\x00\x09\xff\xff\xff\xef']
    ])
    def test_set_response_hops(self, response_hops, dpa):
        request = SetHopsRequest(request_hops=255, response_hops=255, msgid='setHopsTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.response_hops = response_hops
        self.json['data']['req']['param']['responseHops'] = response_hops
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, 255],
        [256, 5],
        [0, -1],
        [239, 1000]
    ])
    def test_construct_invalid(self, request_hops, response_hops):
        with self.assertRaises(ValueError):
            SetHopsRequest(request_hops=request_hops, response_hops=response_hops)



"""Sensor standard response messages."""

from .enumerate import EnumerateResponse
from .read_sensors import ReadSensorsResponse
from .read_sensors_with_types import ReadSensorsWithTypesResponse

__all__ = [
    'EnumerateResponse',
    'ReadSensorsResponse',
    'ReadSensorsWithTypesResponse',
]



import unittest
from iqrfpy.enums.commands import NodeRequestCommands
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.confirmation import Confirmation


class ConfirmationTestCase(unittest.TestCase):

    def test_from_dpa_ok(self):
        message = Confirmation.from_dpa(b'\x01\x00\x01\x00\xff\xff\xff\x36\x01\x04\x01')
        self.assertEqual(message.nadr, 1)
        self.assertEqual(message.pnum, EmbedPeripherals.NODE)
        self.assertEqual(message.pcmd, NodeRequestCommands.READ)
        self.assertEqual(message.hwpid, 65535)
        self.assertEqual(message.dpa_value, 54)
        self.assertEqual(message.request_hops, 1)
        self.assertEqual(message.response_hops, 1)
        self.assertEqual(message.timeslot, 4)

    def test_from_dpa_invalid_len(self):
        with self.assertRaises(ValueError):
            Confirmation.from_dpa(b'\x01\x00\x01\x00\xff\xff')

    def test_from_dpa_not_confirmation(self):
        with self.assertRaises(ValueError):
            Confirmation.from_dpa(b'\x01\x00\x01\x00\xff\xff\x00\x36\x01\x04\x01')

    def test_from_json_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            Confirmation.from_json({})



"""Exploration Peripheral Enumeration Data object."""

from dataclasses import dataclass
from typing import List


@dataclass
class PeripheralEnumerationData:
    """Exploration Peripheral Enumeration Data class."""

    __slots__ = 'dpa_version', 'user_per_nr', 'embedded_pers', 'hwpid', 'hwpid_ver', 'flags', 'user_per'

    def __init__(self, result: dict):
        """Peripheral Enumeration Data constructor.

        Args:
            result (dict): Peripheral enumeration result.
        """
        self.dpa_version: int = result['dpaVer']
        """DPA version."""
        self.user_per_nr: int = result['perNr']
        """Number of non-embedded peripherals implemented by custom DPA handler."""
        self.embedded_pers: List[int] = result['embeddedPers']
        """List of embedded peripherals implemented by device."""
        self.hwpid: int = result['hwpid']
        """Hardware profile ID."""
        self.hwpid_ver: int = result['hwpidVer']
        """Hardware profile version."""
        self.flags: int = result['flags']
        """Device function flags."""
        self.user_per: List[int] = result['userPer']
        """List of non-embedded peripherals implemented by device."""



"""Exploration Peripheral Information response message."""

from typing import List, Optional, Union
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.enums.commands import ExplorationResponseCommands
from iqrfpy.enums.message_types import ExplorationMessages
from iqrfpy.enums.peripherals import Peripheral
from iqrfpy.utils.common import Common
from iqrfpy.utils import dpa as dpa_constants
from iqrfpy.utils.dpa import ResponsePacketMembers, ResponseCodes
from iqrfpy.utils.validators import DpaValidator, JsonValidator
from iqrfpy.peripherals.exploration.responses.peripheral_information_data import PeripheralInformationData

__all__ = [
    'PeripheralInformationResponse',
]


class PeripheralInformationResponse(IResponseGetterMixin):
    """Exploration Peripheral Information response class."""

    __slots__ = ('_peripheral_data',)

    def __init__(self, nadr: int, pnum: Union[Peripheral, int], hwpid: int = dpa_constants.HWPID_MAX,
                 rcode: int = 0, dpa_value: int = 0, msgid: Optional[str] = None, pdata: Optional[List[int]] = None,
                 result: Optional[dict] = None):
        """Read response constructor.

        Args:
            nadr (int): Device address.
            pnum (pnum: Union[Peripheral, int]): Requested peripheral.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            pdata (List[int], optional): DPA response data. Defaults to None.
            msgid (str, optional): Message ID. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pnum=pnum,
            pcmd=ExplorationResponseCommands.PERIPHERALS_ENUMERATION_INFORMATION,
            m_type=ExplorationMessages.PERIPHERAL_INFORMATION,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            msgid=msgid,
            pdata=pdata,
            result=result
        )
        self._peripheral_data = PeripheralInformationData(result=result) if rcode == ResponseCodes.OK else None

    @property
    def peripheral_data(self) -> Optional[PeripheralInformationData]:
        """:obj:`PeripheralInformationData` or :obj:`None`: Peripheral information data.

        Getter only.
        """
        return self._peripheral_data

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'PeripheralInformationResponse':
        """DPA response factory method.

        Parses DPA data and constructs PeripheralInformationResponse object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            PeripheralInformationResponse: Response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        nadr = dpa[ResponsePacketMembers.NADR]
        pnum = dpa[ResponsePacketMembers.PNUM]
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        pdata = None
        result = None
        if rcode == ResponseCodes.OK:
            DpaValidator.response_length(dpa=dpa, expected_len=12)
            pdata = Common.pdata_from_dpa(dpa=dpa)
            result = {
                'perTe': dpa[8],
                'perT': dpa[9],
                'par1': dpa[10],
                'par2': dpa[11],
            }
        return cls(nadr=nadr, pnum=pnum, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata, result=result)

    @classmethod
    def from_json(cls, json: dict) -> 'PeripheralInformationResponse':
        """JSON response factory method.

        Parses JSON API response and constructs PeripheralInformationResponse object.

        Args:
            json (dict): JSON API response.
        Returns:
            PeripheralInformationResponse: Response message object.
        """
        JsonValidator.response_received(json=json)
        nadr = Common.nadr_from_json(json=json)
        pnum = Common.pnum_from_json(json=json)
        msgid = Common.msgid_from_json(json=json)
        hwpid = Common.hwpid_from_json(json=json)
        dpa_value = Common.dpa_value_from_json(json=json)
        rcode = Common.rcode_from_json(json=json)
        pdata = Common.pdata_from_json(json=json)
        result = Common.result_from_json(json=json) if rcode == ResponseCodes.OK else None
        if rcode < 0:
            pnum += 0x80
        return cls(nadr=nadr, pnum=pnum, msgid=msgid, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata,
                   result=result)



"""Request message class abstraction."""

import math
from abc import ABC, abstractmethod
from typing import List, Optional, Union
from uuid import uuid4
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.enums.commands import Command
from iqrfpy.enums.message_types import MessageType
from iqrfpy.enums.peripherals import Peripheral
from iqrfpy.exceptions import RequestNadrInvalidError, RequestPnumInvalidError, RequestPcmdInvalidError, \
    RequestHwpidInvalidError, RequestParameterInvalidValueError

__all__ = ['IRequest']


class IRequest(ABC):
    """This abstract class serves as an interface for embedded peripherals and standards request messages.
    The class provides getters and setters for shared request properties in the form of the @property decorator.
    Shared properties such as device address, peripheral, peripheral command and hwpid are validated here.
    The class also provides generic DPA and JSON serialization methods for derived classes to amend.
    While the abstract class cannot be instantiated, it's methods can be called from a derived class using super.

    Network address (nadr) specifies target device: 0 (coordinator), 1-239 (node), 252 (local device),
    254 (temporary address) and 255 (broadcast), other values are reserved. Values in range 0-255.

    Peripheral number (pnum) specifies peripheral to use: 0-31 are served for embedded peripherals and
    64-127 are reserved for IQRF standard peripherals. Peripheral number can be passed as a member of one
    of the peripheral enum class members (see iqrfpy.enums.peripherals) or plain integer. Values in range 0-255.

    Peripheral command (pcmd) specifies a command to execute, each peripheral implements different commands.
    See DPA framework document for specific command numbers. Peripheral command can be passed as a member of one
    of the peripheral command enum class members (see iqrfpy.enums.commands) or plain integer.
    Peripheral number and peripheral command are pre-defined for each command message this library implements,
    however they will be required for user-defined messages. Values in range 0-255, but note that only values
    0-127 are used for request commands, as response command value is equal to request command value plus 128 (0x80).

    Hardware profile ID (hwpid) specifies a unique device (by its functionality), this argument can be used to filter
    which devices execute a commend specified by request message. Only devices with matching hwpid will execute
    a command, unless the value 65535 (0xFFFF) is used, which ignores hwpid checking. Values in range 0-65535.

    Packet data (pdata) can be required or optional data sent with the request message, depending on the command.
    On a general level, packet data is treated as list of unsigned 1B integers. This means that when defining a user
    command and a request message, request data should be serialized into a list of previously mentioned
    format and stored in the pdata member.

    For more detailed information, see the IQRF DPA Framework documentation.

    Message type (m_type) is an optional argument, however, it is required when using JSON API as message type
    is equivalent to a combination of peripheral number and peripheral command (see iqrfpy.enums.message_types).
    For example, `OSMessages.READ` (iqrfEmbedOs_Read) is equivalent to `EmbedPeripherals.OS` (2) and
    `OSRequestCommands.READ` (0).

    Request parameters (params) is an optional dictionary argument, that is used when constructing a JSON API request.
    When creating a user message, use the params member to propagate your request parameters into the final
    JSON request object.

    DPA response time (dpa_rsp_time) is an optional argument specifying DPA request timeout, not time between reception
    of a request and delivery of response. Note that handling of this argument is the responsibility of a transport
    and system facilitating communication between the library and devices of the IQRF network.

    Device processing time (dev_process_time) is an optional argument used to specify time allotted to the device
    executing a task beyond simple data processing. For example, a device may need to measure some quantity
    for certain amount of time for the measured value to be valid. Note that handling of this argument
    is the responsibility of a transport and system facilitating communication between the library and devices
    of the IQRF network.

    Message ID (msgid) is an optional argument used only by the JSON API at this time. It is used to pair
    JSON API requests and responses as the message type argument only allows to pair requests of the same type,
    but not pairing a specific request and response, given the asynchronous nature of the JSON API.
    """

    __slots__ = '_nadr', '_pnum', '_pcmd', '_mtype', '_hwpid', '_pdata', '_msgid', '_params', '_dpa_rsp_time', \
        '_dev_process_time'

    def __init__(self, nadr: int, pnum: Union[Peripheral, int], pcmd: Union[Command, int],
                 hwpid: int = dpa_constants.HWPID_MAX, pdata: Optional[List[int]] = None,
                 m_type: Optional[MessageType] = None, msgid: Optional[str] = None,
                 params: Optional[dict] = None, dpa_rsp_time: Optional[float] = None,
                 dev_process_time: Optional[float] = None):
        """IRequest constructor.

        Args:
            nadr (int): Network address.
            pnum (Union[Peripheral, int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            pdata (List[int], optional): DPA request data. Defaults to None.
            m_type (MessageType, optional): JSON API message type. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
            params (dict, optional): JSON API request data. Defaults to None.
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
        """
        self._nadr = nadr
        self._pnum = pnum
        self._pcmd = pcmd
        self._hwpid = hwpid
        self._pdata = pdata
        self._mtype = m_type
        self._msgid = msgid if msgid is not None else str(uuid4())
        self._params = params if params is not None else {}
        self._dpa_rsp_time = dpa_rsp_time
        self._dev_process_time = dev_process_time
        self._validate_base(nadr, pnum, pcmd, hwpid, dpa_rsp_time, dev_process_time)

    def _validate_base(self, nadr: int, pnum: Union[Peripheral, int], pcmd: Union[Command, int], hwpid: int,
                       dpa_rsp_time: Optional[float], dev_process_time: Optional[float]) -> None:
        """Shared members validation method.

        Validates shared request members to ensure the data can fit into the DPA packet and JSON request
        object and do not violate rules for expected format of these properties.

        Args:
            nadr (int): Network address.
            pnum (Union[Peripheral, int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
        """
        self._validate_nadr(nadr)
        self._validate_pnum(pnum)
        self._validate_pcmd(pcmd)
        self._validate_hwpid(hwpid)
        self._validate_dpa_rsp_time(dpa_rsp_time)
        self._validate_dev_process_time(dev_process_time)

    @staticmethod
    def _validate_nadr(nadr: int):
        """Validates network address parameter.

        Args:
            nadr (int): Network address value.

        Raises:
            RequestNadrInvalidError: If nadr is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= nadr <= dpa_constants.BYTE_MAX:
            raise RequestNadrInvalidError('NADR should be between 0 and 255.')

    @property
    def nadr(self) -> int:
        """:obj:`int`: Network address.

        Getter and setter.
        """
        return self._nadr

    @nadr.setter
    def nadr(self, value: int):
        self._validate_nadr(value)
        self._nadr = value

    @staticmethod
    def _validate_pnum(pnum: Union[Peripheral, int]):
        """Validates peripheral number parameter.

        Args:
            pnum (Union[Peripheral, int]): Peripheral number value.

        Raises:
            RequestPnumInvalidError: If pnum is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= pnum <= dpa_constants.BYTE_MAX:
            raise RequestPnumInvalidError('PNUM should be between 0 and 255.')

    @property
    def pnum(self) -> Union[Peripheral, int]:
        """:obj:`iqrfpy.enums.peripherals.Peripheral` of :obj:`int`: Peripheral number.

        Getter and setter.
        """
        return self._pnum

    @pnum.setter
    def pnum(self, value: Union[Peripheral, int]):
        self._validate_pnum(pnum=value)
        self._pnum = value

    @staticmethod
    def _validate_pcmd(pcmd: Union[Command, int]):
        """Validates peripheral command parameter.

        Args:
            pcmd (Union[Command, int]): Peripheral command value.

        Raises:
            RequestPcmdInvalidError: If pcmd is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= pcmd <= dpa_constants.BYTE_MAX:
            raise RequestPcmdInvalidError('PCMD should be between 0 and 255.')

    @property
    def pcmd(self) -> Union[Command, int]:
        """:obj:`iqrfpy.enums.commands.Command` or :obj:`int`: Peripheral command.

        Getter and setter.
        """
        return self._pcmd

    @pcmd.setter
    def pcmd(self, value: Union[Command, int]):
        self._validate_pcmd(pcmd=value)
        self._pcmd = value

    @staticmethod
    def _validate_hwpid(hwpid: int):
        """Validates Hardware profile ID parameter.

        Args:
            hwpid (int): Hardware profile ID value.

        Raises:
            RequestHwpidInvalidError: If hwpid is less than 0 or greater than 65535.
        """
        if not dpa_constants.HWPID_MIN <= hwpid <= dpa_constants.HWPID_MAX:
            raise RequestHwpidInvalidError('HWPID should be between 0 and 65535.')

    @property
    def hwpid(self) -> int:
        """:obj:`int`: Hardware profile ID.

        Getter and setter.
        """
        return self._hwpid

    @staticmethod
    def _validate_dpa_rsp_time(dpa_rsp_time: Optional[float] = None):
        """Validates dpa timeout parameter.

        Args:
            dpa_rsp_time (float, optional): DPA timeout value in seconds. Defaults to None.
        Raises:
            RequestParameterInvalidValueError: If dpa_rsp_time is not None and is less than 0.
        """
        if dpa_rsp_time is None:
            return
        if dpa_rsp_time < 0:
            raise RequestParameterInvalidValueError('DPA response time should a positive integer.')

    @property
    def dpa_rsp_time(self) -> Optional[float]:
        """:obj:`float` or :obj:`None`: DPA timeout.

        Getter and setter.
        """
        return self._dpa_rsp_time

    @dpa_rsp_time.setter
    def dpa_rsp_time(self, value: Optional[float] = None):
        self._validate_dpa_rsp_time(value)
        self._dpa_rsp_time = value

    @staticmethod
    def _validate_dev_process_time(dev_process_time: Optional[float] = None):
        """Validates device processing time parameter.

        Args:
            dev_process_time (float, optional): Device processing time value. Defaults to None.
        Raises:
            RequestParameterInvalidValueError: If dev_process_time is not None and is less than 0.
        """
        if dev_process_time is None:
            return
        if dev_process_time < 0:
            raise RequestParameterInvalidValueError('Device processing time should be a positive integer.')

    @property
    def dev_process_time(self) -> Optional[float]:
        """:obj:`float` or :obj:`None`: Device processing time.

        Getter and setter.
        """
        return self._dev_process_time

    @dev_process_time.setter
    def dev_process_time(self, value: Optional[float] = None):
        self._validate_dev_process_time(value)
        self._dev_process_time = value

    @property
    def msgid(self) -> str:
        """:obj:`str`: Message ID.

        Getter and setter.
        """
        return self._msgid

    @msgid.setter
    def msgid(self, value: str):
        self._msgid = value

    @property
    def mtype(self) -> MessageType:
        """:obj:`MessageType`

        Getter only.
        """
        return self._mtype

    @abstractmethod
    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Serializes a request to DPA request packet format. Each request contains the DPA request header.
        If pdata is specified, it is appended to the DPA request header. The request can be serialized
        into both mutable and immutable (default) byte formats.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        dpa: List[int] = [self._nadr, 0, self._pnum, self._pcmd, self._hwpid & 0xFF, (self._hwpid >> 8) & 0xFF]
        if self._pdata is not None:
            dpa.extend(self._pdata)
        if mutable:
            return bytearray(dpa)
        return bytes(dpa)

    @abstractmethod
    def to_json(self) -> dict:
        """JSON API request serialization method.

        Serializes a request to JSON API request object. Each request contains message type and data.
        Data carries message ID for pairing requests with responses, and `req` object that contains network address,
        hwpid and optional request parameters (equivalent to DPA packet data) if `params` is not None.
        If DPA timeout is specified, the value is multiplied by 1000 as JSON API accepts timeout in milliseconds.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        json: dict = {
            'mType': self._mtype.value,
            'data': {
                'msgId': self._msgid,
                'req': {
                    'nAdr': self._nadr,
                    'hwpId': self._hwpid,
                    'param': self._params,
                },
                'returnVerbose': True,
            },
        }
        if self._dpa_rsp_time is not None:
            json['data']['timeout'] = math.ceil(self._dpa_rsp_time * 1000)
        return json



"""RAM Read Any request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import RAMRequestCommands
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['ReadAnyRequest']


class ReadAnyRequest(IRequest):
    """RAM Read Any request class."""

    __slots__ = '_address', '_length'

    def __init__(self, nadr: int, address: int, length: int, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Read Any request constructor.

        Args:
            nadr (int): Device address.
            address (int): Memory address to read from.
            length (int): Number of bytes to read from memory.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(address, length)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.RAM,
            pcmd=RAMRequestCommands.READ_ANY,
            m_type=None,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._address = address
        self._length = length

    def _validate(self, address: int, length: int) -> None:
        """Validate request parameters.

        Args:
            address (int): Memory address to read from.
            length (int): Number of bytes to read from memory.
        """
        self._validate_address(address)
        self._validate_length(length)

    @staticmethod
    def _validate_address(address: int):
        """Validates memory address parameter.

        Args:
            address (int): Memory address to read from.
        Raises:
            RequestParameterInvalidValueError: If address is less than 0 or greater than 65535.
        """
        if not dpa_constants.WORD_MIN <= address <= dpa_constants.WORD_MAX:
            raise RequestParameterInvalidValueError('Address should be between 0 and 65535.')

    @property
    def address(self) -> int:
        """:obj:`int`: Memory address to read from.

        Getter and setter.
        """
        return self._address

    @address.setter
    def address(self, value: int) -> None:
        self._validate_address(address=value)
        self._address = value

    @staticmethod
    def _validate_length(length: int):
        """Validates data length parameter.

        Args:
            length (int): Number of bytes to read from memory.
        Raises:
            RequestParameterInvalidValueError: If length is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= length <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Length should be between 0 and 255.')

    @property
    def length(self) -> int:
        """:obj:`int`: Number of bytes to read from memory.

        Getter and setter.
        """
        return self._length

    @length.setter
    def length(self, value: int) -> None:
        self._validate_length(length=value)
        self._length = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet, if mutable is True.
        """
        self._pdata = [self._address & 0xFF, (self._address >> 8) & 0xFF, self._length]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        This method is not implemented as the message type is not supported by Daemon.

        Returns:
            :obj:`dict`: JSON API request object.
        Raises:
            NotImplementedError: ReadAny message type not implemented.
        """
        raise NotImplementedError('ReadAny JSON API request not implemented.')



"""Sensor Read Sensors With Types request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import SensorRequestCommands
from iqrfpy.enums.message_types import SensorMessages
from iqrfpy.enums.peripherals import Standards
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest
from iqrfpy.peripherals.sensor.requests.sensor_written_data import SensorWrittenData

__all__ = [
    'ReadSensorsWithTypesRequest',
]


class ReadSensorsWithTypesRequest(IRequest):
    """Sensor Read Sensors With Types request class."""

    __slots__ = '_sensors', '_written_data'

    def __init__(self, nadr: int, sensors: Optional[List[int]] = None,
                 written_data: Optional[List[SensorWrittenData]] = None, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Read Sensors With Types request constructor.

        Args:
            nadr (int): Device address.
            sensors (List[int], optional): List of sensor indexes to read. If left empty, only the first sensor is read.
            written_data (List[SensorWrittenData], optional): Data to write to sensors.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_sensors(sensors)
        super().__init__(
            nadr=nadr,
            pnum=Standards.SENSOR,
            pcmd=SensorRequestCommands.READ_SENSORS_WITH_TYPES,
            m_type=SensorMessages.READ_SENSORS_WITH_TYPES,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._sensors = sensors
        self._written_data = written_data

    @staticmethod
    def _validate_sensors(sensors: Optional[List[int]] = None):
        """Validate sensors parameter.

        Args:
            sensors (List[int], optional): List of sensor indexes to read. If left empty, only the first sensor is read.
        Raises:
            RequestParameterInvalidValueError: If sensors contains more than 32 values or if values are not
                in range from 0 to 31.
        """
        if sensors is not None:
            if len(sensors) == 0:
                return
            if len(sensors) > 32:
                raise RequestParameterInvalidValueError('Sensors length should be at most 32 bytes.')
            if min(sensors) < 0 or max(sensors) > 31:
                raise RequestParameterInvalidValueError('Sensors values should be between 0 and 31.')

    @property
    def sensors(self) -> Optional[List[int]]:
        """:obj:`list` of :obj:`int` or :obj:`None`: List of sensor indexes to read.
            If left empty, only the first sensor is read.

        Getter and setter.
        """
        return self._sensors

    @sensors.setter
    def sensors(self, value: Optional[List[int]] = None):
        self._validate_sensors(value)
        self._sensors = value

    @property
    def written_data(self) -> Optional[List[SensorWrittenData]]:
        """:obj:`list` of :obj:`SensorWrittenData` or :obj:`None`: Data to write to sensors.

        Getter and setter.
        """
        return self._written_data

    @written_data.setter
    def written_data(self, value: List[SensorWrittenData]):
        self._written_data = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        if self._sensors is not None:
            written_data = []
            if self._written_data is not None:
                for data in self._written_data:
                    written_data += data.to_pdata()
            self._pdata = Common.indexes_to_4byte_bitmap(self._sensors) + written_data
        else:
            self._pdata = None
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        if self._sensors is not None:
            params = {
                'sensorIndexes': self._sensors
            }
            if self._written_data is not None:
                params['writtenData'] = [data.to_pdata() for data in self._written_data]
            self._params = params
        else:
            self._params = {}
        return super().to_json()



"""Utility enums module.

This module contains auxiliary enums enriched with quality of life functions.
"""

import enum

__all__ = ['IntEnumMember']


class IntEnumMember(enum.IntEnum):
    """Integer enum base class with member value method."""

    @classmethod
    def has_value(cls, value: int) -> bool:
        """Check if enum contains member.

        Args:
            value (int): Value to find in enum

        Returns:
            :obj`bool`: True if value is member of enum, False otherwise
        """
        return value in cls._value2member_map_
